/*
 * =====================================================
 *  _____                _     ____  _____ _   _ _   _ 
 * |_   _|_ __ _   _ ___| |_  / __ \|  ___| | | | | | |
 *   | | | '__| | | / __| __|/ / _` | |_  | |_| | |_| |
 *   | | | |  | |_| \__ \ |_| | (_| |  _| |  _  |  _  |
 *   |_| |_|   \__,_|___/\__|\ \__,_|_|   |_| |_|_| |_|
 *                            \____/                   
 *
 * =====================================================
 *
 * This software is Copyright (C) 2006-2011
 * Fachhochschule Hannover (University of Applied Sciences and Arts)
 * Use is subject to license conditions.
 * 
 * The main licensing options available are:
 * 
 * Open Source Licensing. This is the appropriate option if you want to
 * share the source code of your application with everyone you
 * distribute it to, and you also want to give them the right to share
 * who uses it. If you wish to use TNC@FHH under Open Source Licensing,
 * you must contribute all your source code to the open source
 * community in accordance with the GPL Version 2 when your application
 * is distributed. See http://www.gnu.org/licenses/gpl-2.0.html/ or gpl-2.0.txt
 * 
 * Commercial Licensing. This is the appropriate option if you are
 * creating proprietary applications and you are not prepared to
 * distribute and share the source code of your application.
 * Contact trust@f4-i.fh-hannover.de for details.
 * 
 * http://trust.inform.fh-hannover.de/
 * 
 */

package de.fhhannover.inform.trust.xacml;

import java.io.File;
import java.io.FileNotFoundException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sun.xacml.AbstractPolicy;
import com.sun.xacml.EvaluationCtx;
import com.sun.xacml.UnknownIdentifierException;
import com.sun.xacml.combine.CombiningAlgFactory;
import com.sun.xacml.combine.PolicyCombiningAlgorithm;
import com.sun.xacml.finder.PolicyFinder;
import com.sun.xacml.finder.PolicyFinderModule;
import com.sun.xacml.finder.PolicyFinderResult;
import com.sun.xacml.support.finder.PolicyCollection;
import com.sun.xacml.support.finder.PolicyReader;
import com.sun.xacml.support.finder.StaticPolicyFinderModule;
import com.sun.xacml.support.finder.TopLevelPolicyException;

/**
 * 
 * 
 * @author Bastian Hellmann
 */
public class DynamicFilePolicyModule extends PolicyFinderModule {

	private PolicyCollection policies;

	private File schemaFile = null;

	private File directory;

	private boolean currentlyUpdating;

	private String policyFilenameExtension;

	private long refreshTime;

	// the policy identifier for any policy sets we dynamically create
	private static final String POLICY_ID = "urn:com:sun:xacml:support:finder:dynamic-policy-set";
	private static URI policyId = null;

	// the logger we'll use for all messages
	private static final Logger logger = Logger
			.getLogger(StaticPolicyFinderModule.class.getName());

	public static final long DEFAULT_REFRESH_TIME = 60000;

	static {
		try {
			policyId = new URI(POLICY_ID);
		} catch (Exception e) {
			// this can't actually happen, but just in case...
			if (logger.isLoggable(Level.SEVERE)) {
				logger.log(Level.SEVERE, "couldn't assign default policy id");
			}
		}
	};

	public DynamicFilePolicyModule(String combiningAlg, File policyDirectory,
			String policyFilenameExtension, long refreshTime)
			throws UnknownIdentifierException, URISyntaxException,
			FileNotFoundException {
		PolicyCombiningAlgorithm alg = (PolicyCombiningAlgorithm) (CombiningAlgFactory
				.getInstance().createAlgorithm(new URI(combiningAlg)));

		this.directory = policyDirectory;
		if (!this.directory.isDirectory()) {
			throw new FileNotFoundException();
		}

		this.policyFilenameExtension = policyFilenameExtension;
		this.policies = new PolicyCollection(alg, policyId);
		this.refreshTime = refreshTime;

		String schemaName = System
				.getProperty(PolicyReader.POLICY_SCHEMA_PROPERTY);
		if (schemaName != null) {
			this.schemaFile = new File(schemaName);
		}
	}

	/**
	 * Always returns <code>true</code> since this module does support finding
	 * policies based on context.
	 * 
	 * @return true
	 */
	public boolean isRequestSupported() {
		return true;
	}

	@Override
	public void init(PolicyFinder finder) {
		PolicyReader reader = new PolicyReader(finder, logger, this.schemaFile);
		new FileUpdater(this, this.refreshTime, reader, this.directory,
				this.policyFilenameExtension).start();
	}

	/**
	 * Finds a policy based on a request's context. If more than one policy
	 * matches, then this either returns an error or a new policy wrapping the
	 * multiple policies (depending on which constructor was used to construct
	 * this instance).
	 * 
	 * @param context
	 *            the representation of the request data
	 * 
	 * @return the result of trying to find an applicable policy
	 */
	public synchronized PolicyFinderResult findPolicy(EvaluationCtx context) {
		try {
			while (this.currentlyUpdating) {
				Thread.sleep(100);
			}

			AbstractPolicy policy = this.policies.getPolicy(context);

			if (policy == null) {
				return new PolicyFinderResult();
			} else {
				return new PolicyFinderResult(policy);
			}
		} catch (TopLevelPolicyException tlpe) {
			return new PolicyFinderResult(tlpe.getStatus());
		} catch (InterruptedException e) {
			return new PolicyFinderResult();
		}
	}

	protected synchronized void setCurrentlyUpdating(boolean status) {
		this.currentlyUpdating = status;
	}

	protected synchronized void setPolicies(PolicyCollection policies) {
		this.policies = policies;
	}
}
