/*
 * =====================================================
 *  _____                _     ____  _____ _   _ _   _ 
 * |_   _|_ __ _   _ ___| |_  / __ \|  ___| | | | | | |
 *   | | | '__| | | / __| __|/ / _` | |_  | |_| | |_| |
 *   | | | |  | |_| \__ \ |_| | (_| |  _| |  _  |  _  |
 *   |_| |_|   \__,_|___/\__|\ \__,_|_|   |_| |_|_| |_|
 *                            \____/                   
 *
 * =====================================================
 *
 * This software is Copyright (C) 2006-2011
 * Fachhochschule Hannover (University of Applied Sciences and Arts)
 * Use is subject to license conditions.
 * 
 * The main licensing options available are:
 * 
 * Open Source Licensing. This is the appropriate option if you want to
 * share the source code of your application with everyone you
 * distribute it to, and you also want to give them the right to share
 * who uses it. If you wish to use TNC@FHH under Open Source Licensing,
 * you must contribute all your source code to the open source
 * community in accordance with the GPL Version 2 when your application
 * is distributed. See http://www.gnu.org/licenses/gpl-2.0.html/ or gpl-2.0.txt
 * 
 * Commercial Licensing. This is the appropriate option if you are
 * creating proprietary applications and you are not prepared to
 * distribute and share the source code of your application.
 * Contact trust@f4-i.fh-hannover.de for details.
 * 
 * http://trust.inform.fh-hannover.de/
 * 
 */

/**
 ** License for components that are based on or use parts of the Sun XACML 
 ** implementation. (http://sunxacml.sourceforge.net/)
 **/
/*
 * Copyright 2003-2005 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistribution of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 * 
 *   2. Redistribution in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed or intended for use in
 * the design, construction, operation or maintenance of any nuclear facility.
 */

package de.fhhannover.inform.trust.xacml;

import java.net.URI;
import java.util.Comparator;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.TreeSet;

import com.sun.xacml.AbstractPolicy;
import com.sun.xacml.combine.PolicyCombiningAlgorithm;
import com.sun.xacml.support.finder.PolicyCollection;

/**
 * Based on {@link PolicyCollection} from the Sun XACML implementation.
 * 
 * @author Bastian Hellmann
 */
public class DynamicPolicyCollection extends PolicyCollection {

	// the actual collection of policies
	private HashMap policies;

	// the single instance of the comparator we'll use for managing versions
	private VersionComparator versionComparator = new VersionComparator();

	public DynamicPolicyCollection() {
		super();
		this.policies = new HashMap();
	}

	public DynamicPolicyCollection(PolicyCombiningAlgorithm combiningAlg,
			URI parentPolicyId) {
		super(combiningAlg, parentPolicyId);
		this.policies = new HashMap();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.sun.xacml.support.finder.PolicyCollection#addPolicy(com.sun.xacml
	 * .AbstractPolicy, java.lang.String)
	 */
	public boolean addPolicy(AbstractPolicy policy, String identifier) {
		if (this.policies.containsKey(identifier)) {
			// this identifier is already is use, so see if this version is
			// already in the set
			TreeSet set = (TreeSet) (this.policies.get(identifier));
			set.remove(policy);
			this.policies.put(identifier, set);

			return set.add(policy);
		} else {
			// this identifier isn't already being used, so create a new
			// set in the map for it, and add the policy
			TreeSet set = new TreeSet(this.versionComparator);
			this.policies.put(identifier, set);

			return set.add(policy);
		}
	}

	public boolean removePolicy(AbstractPolicy policy) {
		return removePolicy(policy, policy.getId().toString());
	}

	/**
	 * @param policy
	 * @return
	 */
	public boolean removePolicy(AbstractPolicy policy, String identifier) {
		if (this.policies.containsKey(identifier)) {
			TreeSet set = (TreeSet) (this.policies.get(identifier));
			return set.remove(policy);
		} else {
			return false;
		}
	}

	/**
	 * A <code>Comparator</code> that is used within this class to maintain
	 * ordering amongst different versions of the same policy. Note that it
	 * actually maintains reverse-ordering, since we want to traverse the sets
	 * in decreasing, not increasing order.
	 * <p>
	 * Note that this comparator is only used when there are multiple versions
	 * of the same policy, which in practice will probably happen far less (from
	 * this class' point of view) than additions or fetches.
	 */
	class VersionComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			// we swap the parameters so that sorting goes largest to smallest
			String v1 = ((AbstractPolicy) o2).getVersion();
			String v2 = ((AbstractPolicy) o1).getVersion();

			// do a quick check to see if the strings are equal (note that
			// even if the strings aren't equal, the versions can still
			// be equal)
			if (v1.equals(v2))
				return 0;

			// setup tokenizers, and walk through both strings one set of
			// numeric values at a time
			StringTokenizer tok1 = new StringTokenizer(v1, ".");
			StringTokenizer tok2 = new StringTokenizer(v2, ".");

			while (tok1.hasMoreTokens()) {
				// if there's nothing left in tok2, then v1 is bigger
				if (!tok2.hasMoreTokens())
					return 1;

				// get the next elements in the version, convert to numbers,
				// and compare them (continuing with the loop only if the
				// two values were equal)
				int num1 = Integer.parseInt(tok1.nextToken());
				int num2 = Integer.parseInt(tok2.nextToken());

				if (num1 > num2)
					return 1;

				if (num1 < num2)
					return -1;
			}

			// if there's still something left in tok2, then it's bigger
			if (tok2.hasMoreTokens())
				return -1;

			// if we got here it means both versions had the same number of
			// elements and all the elements were equal, so the versions
			// are in fact equal
			return 0;
		}
	}

}
