/*
 * =====================================================
 *  _____                _     ____  _____ _   _ _   _ 
 * |_   _|_ __ _   _ ___| |_  / __ \|  ___| | | | | | |
 *   | | | '__| | | / __| __|/ / _` | |_  | |_| | |_| |
 *   | | | |  | |_| \__ \ |_| | (_| |  _| |  _  |  _  |
 *   |_| |_|   \__,_|___/\__|\ \__,_|_|   |_| |_|_| |_|
 *                            \____/                   
 *
 * =====================================================
 *
 * This software is Copyright (C) 2006-2011
 * Fachhochschule Hannover (University of Applied Sciences and Arts)
 * Use is subject to license conditions.
 * 
 * The main licensing options available are:
 * 
 * Open Source Licensing. This is the appropriate option if you want to
 * share the source code of your application with everyone you
 * distribute it to, and you also want to give them the right to share
 * who uses it. If you wish to use TNC@FHH under Open Source Licensing,
 * you must contribute all your source code to the open source
 * community in accordance with the GPL Version 2 when your application
 * is distributed. See http://www.gnu.org/licenses/gpl-2.0.html/ or gpl-2.0.txt
 * 
 * Commercial Licensing. This is the appropriate option if you are
 * creating proprietary applications and you are not prepared to
 * distribute and share the source code of your application.
 * Contact trust@f4-i.fh-hannover.de for details.
 * 
 * http://trust.inform.fh-hannover.de/
 * 
 */

package de.fhhannover.inform.trust.xacml;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import com.sun.xacml.AbstractPolicy;
import com.sun.xacml.ParsingException;
import com.sun.xacml.support.finder.PolicyReader;

/**
 * 
 * 
 * @author Bastian Hellmann
 */
public class FileUpdater extends Thread {

	private DynamicFilePolicyModule policyModule;

	private long refreshTime;

	private File directory;

	private Map<File, Long> fileLastModifiedMap;

	private Map<File, AbstractPolicy> filePolicyMap;

	private PolicyReader reader;

	private DynamicPolicyCollection policies;

	private String policyFilenameExtension;

	private static Logger logger = Logger.getLogger(FileUpdater.class);

	public FileUpdater(DynamicFilePolicyModule policyModule, long refreshTime,
			PolicyReader reader, File directory, String policyFilenameExtension) {
		this.policyModule = policyModule;
		this.refreshTime = refreshTime;
		this.reader = reader;
		this.directory = directory;
		this.policyFilenameExtension = policyFilenameExtension;

		this.init();
		this.policyModule.setPolicies(this.policies);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Thread#run()
	 */
	public void run() {
		while (true) {
			try {
				Thread.sleep(this.refreshTime);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}

			logger.debug("Start updating policies");
			this.policyModule.setCurrentlyUpdating(true);
			boolean somethingChanged = this.checkForUpdates();
			if (somethingChanged) {
				this.policyModule.setPolicies(this.policies);
			}
			this.policyModule.setCurrentlyUpdating(false);
			logger.debug("Finished updating policies");
		}
	}

	private void init() {
		this.fileLastModifiedMap = new HashMap<File, Long>();
		this.filePolicyMap = new HashMap<File, AbstractPolicy>();
		this.policies = new DynamicPolicyCollection();
		AbstractPolicy policy = null;

		logger.info("Initializing policies");
		for (File file : this.directory.listFiles()) {
			if (file.isFile()
					&& file.getName().endsWith(this.policyFilenameExtension)) {
				try {
					logger.trace("Try to read policy file: " + file.getName());
					policy = this.reader.readPolicy(file);
					// we loaded the policy, so try putting it in the collection
					if (!this.policies.addPolicy(policy)) {
						logger.warn("tried to load the same policy multiple times: "
								+ file.getName());
					} else {
						logger.trace("Put file & last modified stamp to map.");
						this.fileLastModifiedMap.put(file, file.lastModified());
						this.filePolicyMap.put(file, policy);
					}
				} catch (ParsingException e) {
					logger.warn("Error reading policy: " + file.getName(), e);
				}
			}
		}

		logger.trace("Number of policies: " + this.fileLastModifiedMap.size());
	}

	private boolean checkForUpdates() {
		logger.info("Checking for changes on files");

		boolean somethingChanged = false;
		AbstractPolicy newPolicy = null;

		/**
		 * Check for new or updated files.
		 */
		for (File file : this.directory.listFiles()) {
			if (file.isFile()) {
				Long lastTimeUpdated = this.fileLastModifiedMap.get(file);
				logger.trace("Last modified stamp in map for file '"
						+ file.getName() + "': " + lastTimeUpdated);
				logger.trace("Last modified stamp current for file '"
						+ file.getName() + "': " + file.lastModified());

				if (lastTimeUpdated == null) {
					try {
						newPolicy = this.reader.readPolicy(file);
						if (this.policies.addPolicy(newPolicy)) {
							logger.debug("New policy was added to database.");
							this.fileLastModifiedMap.put(file,
									file.lastModified());
							this.filePolicyMap.put(file, newPolicy);
							somethingChanged = true;
						} else {
							logger.warn("New policy was NOT added to databse.");
						}
					} catch (ParsingException e) {
						e.printStackTrace();
					}
				} else if (lastTimeUpdated < file.lastModified()) {
					try {
						newPolicy = this.reader.readPolicy(file);
						if (this.policies.addPolicy(newPolicy)) {
							logger.debug("Updated version of policy was added to database.");
							this.fileLastModifiedMap.put(file,
									file.lastModified());
							this.filePolicyMap.put(file, newPolicy);
							somethingChanged = true;
						} else {
							logger.warn("Updated version of policy was NOT added to databse; old version remains");
						}
					} catch (ParsingException e) {
						e.printStackTrace();
					}
				}
			}
		}

		/**
		 * Check for deleted files.
		 */
		AbstractPolicy policyToRemove;

		Set<File> keySet = this.filePolicyMap.keySet();
		for (Iterator<File> iterator = keySet.iterator(); iterator.hasNext();) {
			File file = iterator.next();

			if (!file.exists()) {
				logger.debug("File '" + file.getName()
						+ "' doesn't exist anymore and will be removed.");
				policyToRemove = this.filePolicyMap.get(file);
				iterator.remove();
				this.fileLastModifiedMap.remove(file);
				this.policies.removePolicy(policyToRemove);
				somethingChanged = true;
			}
		}

		if (somethingChanged) {
			logger.trace("Number of policies: "
					+ this.fileLastModifiedMap.size());
		} else {
			logger.trace("Nothing changed.");
		}

		return somethingChanged;
	}
}
